package org.greenstone.gatherer.valuetree;
/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.tree.*;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.gui.EditorDialog;
import org.greenstone.gatherer.gui.MetaEditPane;
import org.greenstone.gatherer.gui.SmarterTextArea;
import org.greenstone.gatherer.gui.combobox.GComboBox;
import org.greenstone.gatherer.gui.combobox.GComboBoxModel;
import org.greenstone.gatherer.msm.ElementWrapper;
import org.greenstone.gatherer.msm.Metadata;
import org.greenstone.gatherer.msm.MetadataComboBoxModel;
import org.greenstone.gatherer.msm.MSMAction;
import org.greenstone.gatherer.msm.MSMEvent;
import org.greenstone.gatherer.msm.MSMListener;
import org.greenstone.gatherer.util.Utility;
/**
 * @author John Thompson, Greenstone Digital Library, University of Waikato
 * @version 2.2
 */
public class GValueTree
	 extends JPanel
	 implements ActionListener, FocusListener, MSMListener, ItemListener, KeyListener, TreeSelectionListener {
	 /** Is the metadata currently shown in the edit controls common to all records selected. */
	 private boolean common = false;
	 /** Certain actions are disabled if the metadata selected isn't at file level. */
	 private boolean file_level = false;
	 private boolean ignore = false;
	 private boolean ignore_item_change = false;
	 private CardLayout card_layout;
	 private DefaultListModel list_model;
	 /** The metadata element that is currently selected. */
	 private ElementWrapper selected_metadata_element = null;
	 private GComboBox elements;
	 private MetadataComboBoxModel em;
	 private GValueModel vm;
	 private Hashtable previous_values = new Hashtable();
	 private JButton add;
	 /** This button creates an editor dialog to allow for an expanded area to type in text. */
	 private JButton expand = null;
	 private JButton remove;
	 private JButton restore;
	 private JButton search;
	 private JButton update;
	 private JLabel element_label;
	 private JLabel value_label;
	 private JList list;
	 private JPanel auto_pane;
	 private JPanel button_pane;
	 private JPanel center_pane;
	 private JPanel list_pane;
	 private JPanel north_pane;
	 private JPanel south_pane;
	 private JPanel tree_pane;
	 private JPanel value_pane;
	 private JTextArea auto_message;
	 private JTextField value;
	 private JTree tree;
	 private MetaEditPane metaedit;
	 private String args[] = new String[1];
	 private TreePath previous;
	 private ValueListener value_listener;
	 /** Stock standard size for labels. */
	 static final private Dimension LABEL_SIZE = new Dimension(132, 26);

	 static final private String LIST = "List";
	 static final private String NONE = "None";
	 static final private String TREE = "Tree";


	 public GValueTree(MetaEditPane metaedit, int width, int height, JButton add, JButton update, JButton remove) {
		  this(metaedit, new Dimension(width, height), add, update, remove);
	 }

	 public GValueTree(MetaEditPane metaedit, Dimension size, JButton add, JButton update, JButton remove) {
		 super();

		 this.add = add;
		 this.metaedit = metaedit;
		 this.remove = remove;
		 this.update = update;

		 ///ystem.err.println("Created new ValueListener!");
		 value_listener = new ValueListener();

		 int height = size.height;
		 int width = size.width;

		 setFont(Gatherer.config.getFont("general.font", false));
		 //setSize(size);
		 setPreferredSize(size);

		 north_pane = new JPanel();

		 element_label = new JLabel(get("Element"));
		 element_label.setPreferredSize(LABEL_SIZE);

		 elements = new GComboBox();
		 elements.addItemListener(this);
		 elements.setEditable(false);
		 elements.setMaximumRowCount(5);
		 elements.setPreferredSize(new Dimension(413, 24));

		 center_pane = new JPanel();

		 card_layout = new CardLayout();

		 auto_pane = new JPanel();

		 args[0] = "Title";
		 auto_message = new SmarterTextArea(get("AutoMessage", args), false);
		 auto_message.setLineWrap(true);
		 auto_message.setOpaque(false);
		 auto_message.setWrapStyleWord(true);

		 tree_pane = new JPanel();

		 JLabel tree_label = new JLabel(get("Tree"));

		 tree = new JTree(new GValueModel());
		 tree.addTreeSelectionListener(this);
		 tree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
		 tree.setRootVisible(true);
		 tree.putClientProperty("JTree.lineStyle", "Angled");
		 //setTreeEnabled(false);

		 list_pane = new JPanel();

		 JLabel list_label = new JLabel(get("List"));

		 list_model = new DefaultListModel();
		 list = new JList(list_model);
		 list.addListSelectionListener(new ListSelectionListenerImpl());

		 south_pane = new JPanel();

		 value_pane = new JPanel();

		 value_label = new JLabel(get("Value"));
		 value_label.setPreferredSize(LABEL_SIZE);

		 value = new JTextField();
		 value.addActionListener(this);
		 value.addFocusListener(this);
		 //value.addKeyListener(this);
		 value.getDocument().addDocumentListener(new DocumentListenerImpl());
		 value.setPreferredSize(new Dimension(413, 24));

		 button_pane = new JPanel();

		 expand = new JButton(get("General.Expand"));
		 expand.addActionListener(this);
		 expand.setMnemonic(KeyEvent.VK_E);

		 restore = new JButton(get("Restore"));
		 restore.addActionListener(this);
		 restore.setEnabled(false);
		 restore.setMnemonic(KeyEvent.VK_T);

		 search = new JButton(get("Search"));
		 search.addActionListener(this);
		 search.setMnemonic(KeyEvent.VK_S);

		 // Layout
		 element_label.setBorder(BorderFactory.createEmptyBorder(0,0,0,5));

		 north_pane.setLayout(new BorderLayout());
		 north_pane.add(element_label, BorderLayout.CENTER);
		 north_pane.add(elements, BorderLayout.EAST);

		 tree_pane.setLayout(new BorderLayout());
		 tree_pane.add(tree_label, BorderLayout.NORTH);
		 tree_pane.add(new JScrollPane(tree), BorderLayout.CENTER);

		 list_pane.setLayout(new BorderLayout());
		 list_pane.add(list_label, BorderLayout.NORTH);
		 list_pane.add(new JScrollPane(list), BorderLayout.CENTER);

		 auto_pane.setBorder(BorderFactory.createEmptyBorder(25,10,25,10));
		 auto_pane.setLayout(new BorderLayout());
		 auto_pane.add(auto_message, BorderLayout.CENTER);

		 center_pane.setBorder(BorderFactory.createEmptyBorder(2,0,2,0));
		 center_pane.setLayout(card_layout);
		 center_pane.add(tree_pane, TREE);
		 center_pane.add(list_pane, LIST);
		 center_pane.add(auto_pane, NONE);

		 value_label.setBorder(BorderFactory.createEmptyBorder(0,0,0,5));

		 value_pane.setBorder(BorderFactory.createEmptyBorder(0,0,2,0));
		 value_pane.setLayout(new BorderLayout());
		 value_pane.add(value_label, BorderLayout.CENTER);
		 value_pane.add(value, BorderLayout.EAST);

		 button_pane.setLayout(new GridLayout(2,3,2,0));
		 button_pane.add(search);
		 button_pane.add(restore);
		 button_pane.add(expand);
		 button_pane.add(add);
		 button_pane.add(update);
		 button_pane.add(remove);

		 south_pane.setLayout(new BorderLayout());
		 south_pane.add(value_pane, BorderLayout.CENTER);
		 south_pane.add(button_pane, BorderLayout.SOUTH);

		 this.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
		 this.setLayout(new BorderLayout());
		 this.add(north_pane, BorderLayout.NORTH);
		 this.add(center_pane, BorderLayout.CENTER);
		 this.add(south_pane, BorderLayout.SOUTH);
	 }

	public void actionPerformed(ActionEvent event) {
		 ignore = true;
		 if(event.getSource() == value) {
			  StringTokenizer tokenizer = new StringTokenizer(value.getText(), "\\");
			  TreePath closest = getClosestPath(tokenizer);
			  if(closest != null) {
					TreePath path = tree.getSelectionPath();
					GValueNode node = (GValueNode)path.getLastPathComponent();
					///ystem.err.println("1. Set value: " + node.toString());
					value.setText(Utility.stripNL(node.toString()));//formatPath(path, false)));
					// Now systematically add to selected record.
					add.doClick();
			  }
			  else {
					int response = JOptionPane.showConfirmDialog(this, "Do you wish to create a new entry?", "Warning", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
					if(response == JOptionPane.YES_OPTION) {
						 TreePath new_path = addNode(value.getText());
						 tree.setSelectionPath(new_path);
						 tree.scrollPathToVisible(new_path);
						 GValueNode node = (GValueNode)new_path.getLastPathComponent();
						 ///ystem.err.println("2. Set value: " + node.toString());
						 value.setText(Utility.stripNL(node.toString()));		
						 // Now systematically add to selected record.
						 add.doClick();
					}
			  }
		 }
		 else if(event.getSource() == search) {
			  searchValues();
			  restore.setEnabled(true);
			  card_layout.show(center_pane, LIST);
		 }
		 else if(event.getSource() == restore) {
			  restore.setEnabled(false);
			  card_layout.show(center_pane, TREE);
		 }
		 else if(event.getSource() == expand) {
			  EditorDialog ed = new EditorDialog();
			  String temp = ed.display(value.getText());
			  if(temp != null) {
					///ystem.err.println("3. Set value: " + temp);
					value.setText(temp);
			  }
			  //validateControls();
		 }
		 ignore = false;
	}
	 /** Called whenever an element within the metadata set module changes in some way.
	  * @param event A <strong>MSMEvent</strong> which contains information about the change.
	  */
	 public void elementChanged(MSMEvent event) {
		  /* @TODO - Not sure. Figure out. */
	 }
	 public void focusGained(FocusEvent event) {

	 }

	 public void focusLost(FocusEvent event) {

	 }

	 public boolean getCommon() {
		  return common;
	 }

	 public ElementWrapper getSelectedMetadataElement() {
		  return selected_metadata_element;
	 }

	 public String getSelectedValue() {
		  String value_text = value.getText();
		  previous_values.put(selected_metadata_element, value_text);
		  return value_text;
	 }

	 public GValueModel getValueModel() {
		  return (GValueModel) tree.getModel();
	 }

	 /** This method is required for any ItemListener so that it can be informed when the state of an item set changes (such as when a new selection is made in a combobox). Note that this update is subtly different from the validateControls() method as it also gathers new models for each combobox as required. The reason this isn't done as part of validateControls() is that this would continously reset whatever the user is attempting to type (ie value would be found to be enabled and a new model would be loaded erasing whatever the user just typed in value to fire the validateControls() in the first place!)
	  * @param event An ItemEvent encapsulating all the information about this event.
	  */
    public void itemStateChanged(ItemEvent event) {
		  ///ystem.err.println("Item State Changed");
		  if(!ignore_item_change) {
				ignore_item_change = true;
				selected_metadata_element = (ElementWrapper)em.getSelectedItem();
				if(selected_metadata_element != null) {
					 elements.setToolTipText(selected_metadata_element.getToolTip());
					 vm = null;
					 vm = Gatherer.c_man.getCollection().msm.getValueTree(selected_metadata_element);
					 if(vm != null) {
						  ///ystem.err.println("Assigning value model.");
						  vm.addTreeModelListener(value_listener);
						  tree.addTreeSelectionListener(this);
						  setTreeEnabled(true);
					 }
					 else {
						  ///ystem.err.println("Creating new model.");
						  vm = new GValueModel(get("Not_Applicable"));
						  setTreeEnabled(false);
					 }
					 tree.setModel(vm);
				}
				else {
					 setTreeEnabled(false);
				}
				value.setText("");
				metaedit.validateControls();
		  }
		  previous = null;
		  ignore_item_change = false;
		  validateControls();
    }

	 public void keyPressed(KeyEvent event) {
	 }

    public void keyReleased(KeyEvent event) {
		  ignore = true;
		  ///ystem.err.println("Key released.");
		  StringTokenizer tokenizer = new StringTokenizer(value.getText(), "\\");
		  TreePath closest = getClosestPath(tokenizer);
		  if(closest == previous) {
				validateControls();
				ignore = false;
				return;
		  }
		  else {
				// Clear previous
				if(previous != null) {
					 // We have a brand new match close up the old one.
					 if(closest != null) {
						  if(getParent(previous) != null &&
							  !getParent(previous).toString().equals
							  ("[AssignedValues]")) {
								tree.collapsePath(getParent(previous));
						  }
						  previous = null;
					 }
					 // We haven't got a match, but we did have a matching path to
					 // this point. Move the selection to previous paths parent.
					 else {
						  tree.setSelectionPath(getParent(previous));
					 }
				}
				
				if(closest != null) {
					 tree.setSelectionPath(closest);
					 tree.scrollPathToVisible(closest);
					 previous = closest;
				}
		  }
		  ignore = false;
		  validateControls();
    }
    
    public void keyTyped(KeyEvent event) {
    }

	 /** Called whenever a metadata value is changed.
	  * @param event A <strong>MSMEvent</strong> which encapsulates relevant data about the change.
	  */
	 public void metadataChanged(MSMEvent event) {
		  // Don't care.
	 }
	 
    public void newMetadataSetManager() {
		  if(Gatherer.c_man != null && Gatherer.c_man.getCollection() != null && Gatherer.c_man.getCollection().msm != null) {
				Gatherer.c_man.getCollection().msm.addMSMListener(this);
		  }
		  setChanged(null);
    }

	 public void refreshValueTree() {
		  if(em != null && em.getSize() > 0) {
				setSelectedMetadataElement((ElementWrapper)em.getSelectedItem());
		  }
	 }
    
    public void registerModel(GValueModel model) {
		  vm = model;
		  vm.addTreeModelListener(value_listener);
		  tree.setModel(model);
		  tree.addTreeSelectionListener(this);
    }
	 /** Called whenever a set within the metadata set module changes in some way.
	  * @param event A <strong>MSMEvent</strong> which contains information about the change.
	  */
	 public void setChanged(MSMEvent event) {
		  // Update the element model.
		  MetadataComboBoxModel model = Gatherer.c_man.getCollection().msm.getElementModel();
		  if(model.getSize() > 0) {
				setElementModel(model);
				model.setSelectedItem(model.getElementAt(0));
				vm = Gatherer.c_man.getCollection().msm.getValueTree((ElementWrapper)model.getSelectedItem());
				setTreeEnabled(true);
				if(vm == null) {
					 vm = new GValueModel(get("Not_Applicable"));
					 setTreeEnabled(false);
				}
		  }
		  else {
				vm = new GValueModel();
				setTreeEnabled(true);
		  }
		  vm.addTreeModelListener(value_listener);
		  tree.setModel(vm);
		  tree.addTreeSelectionListener(this);
	 } 

	 public void setCommon(boolean common) {
		  this.common = common;
	 }

    public void setElementModel(MetadataComboBoxModel model) {
		  ///ystem.err.println("Reset ElementModel.");
		  setTreeEnabled(true);
		  Gatherer.c_man.getCollection().msm.addMSMListener(this);
		  if(model == null) {
				Gatherer.println("model = null");
				System.exit(0);
		  }
		  ignore_item_change = true;
		  em = model;
		  elements.setModel(em);
		  if(em.getSize() > 0) {
				///ystem.err.println("em.size() > 0");
				elements.setEnabled(true);
				elements.setSelectedIndex(0);
				selected_metadata_element = (ElementWrapper)em.getSelectedItem();
				elements.setToolTipText(selected_metadata_element.getToolTip());
				// First we create and install a blank model, or else no updates ever work.
				vm = new GValueModel(get("Not_Applicable"));
				tree.setModel(vm);
				// Now we create a new model if applicable, disabling otherwise.
				if(selected_metadata_element.getNamespace().length() > 0) {
					 vm = Gatherer.c_man.getCollection().msm.getValueTree(selected_metadata_element);
					 vm.addTreeModelListener(value_listener);
					 ///ystem.err.println("Found existing model containing " + vm.size() + " entries");
					 setTreeEnabled(true);
					 tree.setModel(vm); // This doesn't seem to work
				}
				else {
					 ///ystem.err.println("Creating new model.");
					 setTreeEnabled(false);
				}
				///ystem.err.println("The value tree should have been updated by now.");
				///ystem.err.println("6. Set value: " + getLastValue());
				value.setText(Utility.stripNL(getLastValue()));
		  }
		  else {
				///ystem.err.println("No elements loaded.");
				elements.setEnabled(false);
				elements.setToolTipText(get("No_Description"));
				//setTreeEnabled(false);
				value.setText("");
		  }
		  previous = null;
		  ignore_item_change = false;
		  validateControls();
    }

	 public void setFileLevel(boolean file_level) {
		  this.file_level = file_level;
	 }

	 public void setRootVisible(boolean val) {
		  tree.setRootVisible(val);
	 }

    public void setSelectedMetadataElement(ElementWrapper element) {
		  if(!ignore_item_change) {
				ignore_item_change = true;
				if(element != null) {
					 selected_metadata_element = element;
					 elements.setSelectedItem(element);
					 elements.setToolTipText(element.getToolTip());
					 vm = Gatherer.c_man.getCollection().msm.getValueTree(element);
					 if(vm != null) {
						  vm.addTreeModelListener(value_listener);
						  setTreeEnabled(true);
					 }
					 else {
						  vm = new GValueModel(get("Not_Applicable"));
						  setTreeEnabled(false);
					 }
					 tree.setModel(vm);
					 ///ystem.err.println("7. Set value: " + getLastValue());
					 value.setText(Utility.stripNL(getLastValue()));
				}
				else {
					 elements.setToolTipText(get("No_Description"));
					 setTreeEnabled(false);
					 value.setText("");
				}
				metaedit.validateControls();
		  }
		  previous = null;
		  ignore_item_change = false;
		  validateControls();
    }

	 public void setSelectedValue(String val) {
		  ignore = true;
		  value.setText(Utility.stripNL(val));
		  StringTokenizer tokenizer = new StringTokenizer(val, "\\");
		  TreePath closest = getClosestPath(tokenizer);
		  if(previous != null) {
				tree.collapsePath(getParent(previous));
				previous = null;
		  }
		  if(closest != null) {
				tree.setSelectionPath(closest);
				tree.scrollPathToVisible(closest);
		  }
		  ignore = false;
		  validateControls();
	 }

	 /** This method enables or diables the various controls based upon the
	  * state of the system. This method should be called after every change
	  * which could affect the GUI.
	  */
	 public void validateControls() {
		  //gatherer.debug("Validating controls...");
		  ///ystem.err.println("Validating controls...");
		  // Validate add
		  if(metaedit.getSelectedNode() != null && value.getText().length() > 0) {
		      add.setEnabled(true);
		      //gatherer.debug("Accepted.");
		      ///ystem.err.println("Yes");
		  }
		  else {
		      add.setEnabled(false);
		      //gatherer.debug("Denied.");
		      ///ystem.err.println("No.");
		  }

		  // Validate update
		  Metadata selected_metadata = metaedit.getSelectedMetadata();
		  if(metaedit.getSelectedNode() != null &&  selected_metadata != null && selected_metadata.getValueNode() != null && selected_metadata.getValueNode().getFullPath().length() > 0 && value.getText().length() > 0 && common) {
				update.setEnabled(true);
		  }
		  else {
				update.setEnabled(false);
		  }

		  // Validate remove
		  if(metaedit.getSelectedNode() != null && metaedit.getSelectedMetadata() != null && file_level) {
				remove.setEnabled(true);
		  }
		  else {
				remove.setEnabled(false);
		  }
	 }

	 /** Called when the value tree of a certain element has changed significantly. If we are displaying the tree for the element noted in the event, then reload the value model.
	  * @param event A <strong>MSMEvent</strong> containing information about the event.
	  */
	 public void valueChanged(MSMEvent event) {
		  ElementWrapper element = event.getElement();
		  if(selected_metadata_element.equals(element)) {
				vm = new GValueModel(get("Not_Applicable"));
				tree.setModel(vm);
				GValueModel temp = Gatherer.c_man.getCollection().msm.getValueTree(element);
				if(temp != null) {
					 vm = temp;
					 tree.setModel(vm);
					 vm.addTreeModelListener(value_listener);
					 setTreeEnabled(true);
				}
				else {
					 setTreeEnabled(false);
				}
		  }
	 }

    public void valueChanged(TreeSelectionEvent event) {
		  if(tree.getSelectionCount() != 0 && !ignore) {
				TreePath path = tree.getSelectionPath();
				GValueNode node = (GValueNode) path.getLastPathComponent();
				value.setText(node.getFullPath());//formatPath(tree.getSelectionPath(), false)));
		  }
		  validateControls();
    }
    
    private TreePath addNode(String reference) {
		  StringTokenizer tokenizer = new StringTokenizer(reference, "\\");
		  GValueModel model = (GValueModel)tree.getModel();
		  GValueNode current = (GValueNode)model.getRoot();
		  while(tokenizer.hasMoreElements()) {
				String token = tokenizer.nextToken();
				GValueNode node = (GValueNode)getClosestNode(token, current);
				if(node == null) {
					 Gatherer.println("Before: '" + current.toString() + "'");
					 for(int i = 0; i < current.getChildCount(); i++) {
						  Gatherer.println(current.getChildAt(i).toString());
					 }
					 node = model.addValue(token, current, null);
					 Gatherer.println("After: '" + current.toString() + "'");
					 for(int i = 0; i < current.getChildCount(); i++) {
						  Gatherer.println(current.getChildAt(i).toString());
					 }
				}
				current = node;
		  }
		  tree.updateUI();
		  if(current != null) {
				return new TreePath(current.getPath());
		  }
		  else {
				return new TreePath(((GValueNode)model.getRoot()).getPath());
		  }
    }
	 /** Returns the given tree path as path formatted string (ie subject\subject\subject).
	  * @param tree The <strong>JTree</strong> the TreePath came from. Used to determine if the root node should be encoded as well.
	  * @param path A <strong>TreePath</strong> that you wish to encode to String.
	  * @param full <i>true</i> if this is the full path (ie a leaf node), <i>false</i> otherwise.
	  * @return A <strong>String</strong> encoding the information from the TreePath.
	  * @see javax.swing.JTree
	  * @see javax.swing.tree.TreePath
	  */
	static public String formatPath(JTree tree, TreePath path, boolean full) {
		String text = "";
		int i = 0;
		if(tree == null || (!tree.isRootVisible() && !full)) {
			 i = 1;
		}
		for( ; i < path.getPathCount(); i++) {
			GValueNode node = (GValueNode)path.getPathComponent(i);
		  text = text + node.toString();
			if(node.getChildCount() > 0) {
				text = text + "\\";
			}
		}
		if(full && text.endsWith("\\")) {
			 return text.substring(0, text.length() - 1);
		}
		return text;
	}

	 private String get(String key) {
		  return get(key, null);
	 }

	 private String get(String key, String args[]) {
		  if(key.indexOf('.') == -1) {
				key = "MetaEdit." + key;
		  }
		  return Gatherer.dictionary.get(key,args);
	 }

	private GValueNode getClosestNode(String target, GValueNode current) {
		// We have struck it lucky.
		if(startsWith(current.toString(), target)) {
			return current;
		}
		GValueNode closest = null;
		for(int i = 0; i < current.getChildCount() && closest == null; i++) {
			closest = getClosestNode(target, (GValueNode)current.getChildAt(i));
		}
		return closest;
	}

	 private TreePath getClosestPath(StringTokenizer tokenizer) {
		  if(tokenizer.countTokens() > 0) {
				TreeNode tree_node = (TreeNode)tree.getModel().getRoot();
				if(tree_node instanceof GValueNode) {
					 GValueNode current = (GValueNode)tree_node;
					 TreePath closest = null;
					 while(tokenizer.hasMoreElements() && current != null) {
						  current = getClosestNode(tokenizer.nextToken(), current);
					 }
					 if(current != null) {
						  return new TreePath(current.getPath());
					 }
					 return null;
				}
		  }
		  return null;
	 }

	 private String getLastValue() {
		  if(previous_values.containsKey(selected_metadata_element)) {
				return (String)previous_values.get(selected_metadata_element);
		  }
		  return "";
	 }

	private TreePath getParent(TreePath path) {
		if(path.getPathCount() > 1) {
			TreeNode nodes[] = new TreeNode[path.getPathCount() - 1];
			for(int i = 0; i < path.getPathCount() - 1; i++) {
				nodes[i] = (TreeNode)path.getPathComponent(i);
			}
			return new TreePath(nodes);
		}
		return null;
	}

	private void searchValues() {
		 // Clear any old search results.
		 list_model.clear();
		 String target = value.getText();
		 GValueNode current = (GValueNode)tree.getModel().getRoot();
		 Vector search_space = new Vector();
		 search_space.add(current);
		 for(int i = 0; i < search_space.size(); i++) {
			  GValueNode node = (GValueNode)search_space.get(i);
			  // Add this nodes children.
			  for(int j = 0; j < node.getChildCount(); j++) {
					search_space.add(node.getChildAt(j));
			  }
			  // And if this node matches add it to list, except for the very
			  // first node as that is AssignedValues the root node!
			  if(node.toString().indexOf(target) != -1 &&
				  node != current) {
					list_model.add(list_model.size(), node);
			  }
		 }
	}

	private void setTreeEnabled(boolean value) {
		 tree.setRootVisible(!value);
		 tree.setEnabled(value);
		 if(value) {
			  card_layout.show(center_pane, TREE);
		 }
		 else {
			  args[0] = selected_metadata_element.toString();
			  auto_message.setText(get("AutoMessage", args));
			  card_layout.show(center_pane, NONE);
		 }
	}

	private boolean startsWith(String longer, String shorter) {
		if(longer.length() > shorter.length()) {
		  longer = longer.substring(0, shorter.length());
		}
		return longer.equalsIgnoreCase(shorter);
	}

    private void printTreePaths() {
	///ystem.out.println("Tree Paths according to getPath() method.");
	Vector nodes = new Vector();
	nodes.add(tree.getModel().getRoot());
		  while(nodes.size() > 0) {
		      GValueNode node = (GValueNode)nodes.get(0);
		      nodes.remove(node);
		      for(int i = 0; i < node.getChildCount(); i++) {
			  nodes.add(node.getChildAt(i));
		      }
				///ystem.out.println("TreePath = " + node.getPath());
		  }
    }

	 private class DocumentListenerImpl
		  implements DocumentListener {
		  /** Gives notification that an attribute or set of attributes changed. */
		  public void changedUpdate(DocumentEvent e) {
				validateControls();
		  }
          
		  /** Gives notification that there was an insert into the document. */
		  public void insertUpdate(DocumentEvent e) {
				validateControls();
		  }
          
		  /** Gives notification that a portion of the document has been removed. */
		  public void removeUpdate(DocumentEvent e) {
				validateControls();
		  }
	 }

	 private class ListSelectionListenerImpl
		  implements ListSelectionListener {
		  public void valueChanged(ListSelectionEvent event) {
				GValueNode node = (GValueNode)list.getSelectedValue();
				value.setText(Utility.stripNL(node.toString()));//formatPath(node.getPath(), false)));
				value.setCaretPosition(0);
		  }
	 }

    private class ValueListener
		  implements TreeModelListener {
		  /* Called when nodes have been modified, but the model hasn't changed.
			* @param event Everything you ever wanted to know about model changes
			* in one tidy TreeModelEvent package.
			*/
		  public void treeNodesChanged(TreeModelEvent event) {
				update();
		  }
		  
		  /* Called when nodes have been added to the model.
			* @param event Everything you ever wanted to know about model changes
			* in one tidy TreeModelEvent package.
			*/
		  public void treeNodesInserted(TreeModelEvent event) {
				update();
		  }
		  
		  /* Called when nodes have been removed from the model.
			* @param event Everything you ever wanted to know about model changes
			* in one tidy TreeModelEvent package.
			*/
		  public void treeNodesRemoved(TreeModelEvent event) {
				update();
		  }
		  
		  /** Called when nodes have been rearranged within the model.
			* @param event Everything you ever wanted to know about model changes
			* in one tidy TreeModelEvent package.
			*/
		  public void treeStructureChanged(TreeModelEvent event) {
				update();
		  }
		  
		  private void update() {
				// Don't bother updating if the metaedit tab isn't selected.
				if(Gatherer.g_man.getSelectedView() == Gatherer.g_man.metaedit_pane) {
					 vm = null;
					 vm = Gatherer.c_man.getCollection().msm.getValueTree((ElementWrapper)em.getSelectedItem());
					 if(vm != null) {
						  vm.addTreeModelListener(value_listener);
					 }
					 else {
						  vm = new GValueModel(get("Not_Applicable"));
					 }
					 tree.setModel(vm);
					 tree.repaint();
				}
		  }
    }
}





